# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("uxtheme.dll")
prototypes = \
    {
        #
        'BeginPanningFeedback': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd"]),
        #
        'UpdatePanningFeedback': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "lTotalOverpanOffsetX", "lTotalOverpanOffsetY", "fInInertia"]),
        #
        'EndPanningFeedback': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "fAnimateBack"]),
        #
        'GetThemeAnimationProperty': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="TA_PROPERTY"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iStoryboardId", "iTargetId", "eProperty", "pvProperty", "cbSize", "pcbSizeOut"]),
        #
        'GetThemeAnimationTransform': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("TA_TRANSFORM", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iStoryboardId", "iTargetId", "dwTransformIndex", "pTransform", "cbSize", "pcbSizeOut"]),
        #
        'GetThemeTimingFunction': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("TA_TIMINGFUNCTION", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iTimingFunctionId", "pTimingFunction", "cbSize", "pcbSizeOut"]),
        #
        'OpenThemeData': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd", "pszClassList"]),
        #
        'OpenThemeDataEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="OPEN_THEME_DATA_FLAGS")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd", "pszClassList", "dwFlags"]),
        #
        'CloseThemeData': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme"]),
        #
        'DrawThemeBackground': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pRect", "pClipRect"]),
        #
        'DrawThemeBackgroundEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("DTBGOPTS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pRect", "pOptions"]),
        #
        'DrawThemeText': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="DRAW_TEXT_FORMAT"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pszText", "cchText", "dwTextFlags", "dwTextFlags2", "pRect"]),
        #
        'GetThemeBackgroundContentRect': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pBoundingRect", "pContentRect"]),
        #
        'GetThemeBackgroundExtent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pContentRect", "pExtentRect"]),
        #
        'GetThemeBackgroundRegion': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pRect", "pRegion"]),
        #
        'GetThemePartSize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypeInt(signed=False, label="THEMESIZE"), SimTypePointer(SimTypeRef("SIZE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "prc", "eSize", "psz"]),
        #
        'GetThemeTextExtent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="DRAW_TEXT_FORMAT"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pszText", "cchCharCount", "dwTextFlags", "pBoundingRect", "pExtentRect"]),
        #
        'GetThemeTextMetrics': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("TEXTMETRICW", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "ptm"]),
        #
        'HitTestThemeBackground': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="HIT_TEST_BACKGROUND_OPTIONS"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeRef("POINT", SimStruct), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "dwOptions", "pRect", "hrgn", "ptTest", "pwHitTestCode"]),
        #
        'DrawThemeEdge': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypeInt(signed=False, label="DRAWEDGE_FLAGS"), SimTypeInt(signed=False, label="DRAW_EDGE_FLAGS"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pDestRect", "uEdge", "uFlags", "pContentRect"]),
        #
        'DrawThemeIcon': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pRect", "himl", "iImageIndex"]),
        #
        'IsThemePartDefined': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId"]),
        #
        'IsThemeBackgroundPartiallyTransparent': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId"]),
        #
        'GetThemeColor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pColor"]),
        #
        'GetThemeMetric': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "iPropId", "piVal"]),
        #
        'GetThemeString': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pszBuff", "cchMaxBuffChars"]),
        #
        'GetThemeBool': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pfVal"]),
        #
        'GetThemeInt': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "piVal"]),
        #
        'GetThemeEnumValue': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "piVal"]),
        #
        'GetThemePosition': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("POINT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pPoint"]),
        #
        'GetThemeFont': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("LOGFONTW", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "iPropId", "pFont"]),
        #
        'GetThemeRect': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pRect"]),
        #
        'GetThemeMargins': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("MARGINS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "iPropId", "prc", "pMargins"]),
        #
        'GetThemeIntList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("INTLIST", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pIntList"]),
        #
        'GetThemePropertyOrigin': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="PROPERTYORIGIN"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pOrigin"]),
        #
        'SetWindowTheme': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "pszSubAppName", "pszSubIdList"]),
        #
        'GetThemeFilename': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "pszThemeFileName", "cchMaxBuffChars"]),
        #
        'GetThemeSysColor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hTheme", "iColorId"]),
        #
        'GetThemeSysColorBrush': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hTheme", "iColorId"]),
        #
        'GetThemeSysBool': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iBoolId"]),
        #
        'GetThemeSysSize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iSizeId"]),
        #
        'GetThemeSysFont': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeRef("LOGFONTW", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iFontId", "plf"]),
        #
        'GetThemeSysString': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iStringId", "pszStringBuff", "cchMaxStringChars"]),
        #
        'GetThemeSysInt': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iIntId", "piValue"]),
        #
        'IsThemeActive': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'IsAppThemed': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'GetWindowTheme': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd"]),
        #
        'EnableThemeDialogTexture': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "dwFlags"]),
        #
        'IsThemeDialogTextureEnabled': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd"]),
        #
        'GetThemeAppProperties': SimTypeFunction([], SimTypeInt(signed=False, label="SET_THEME_APP_PROPERTIES_FLAGS")),
        #
        'SetThemeAppProperties': SimTypeFunction([SimTypeInt(signed=False, label="SET_THEME_APP_PROPERTIES_FLAGS")], SimTypeBottom(label="Void"), arg_names=["dwFlags"]),
        #
        'GetCurrentThemeName': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pszThemeFileName", "cchMaxNameChars", "pszColorBuff", "cchMaxColorChars", "pszSizeBuff", "cchMaxSizeChars"]),
        #
        'GetThemeDocumentationProperty': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pszThemeName", "pszPropertyName", "pszValueBuff", "cchMaxValChars"]),
        #
        'DrawThemeParentBackground': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "hdc", "prc"]),
        #
        'EnableTheming': SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["fEnable"]),
        #
        'DrawThemeParentBackgroundEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="DRAW_THEME_PARENT_BACKGROUND_FLAGS"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "hdc", "dwFlags", "prc"]),
        #
        'SetWindowThemeAttribute': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="WINDOWTHEMEATTRIBUTETYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "eAttribute", "pvAttribute", "cbAttribute"]),
        #
        'DrawThemeTextEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="DRAW_TEXT_FORMAT"), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypePointer(SimTypeRef("DTTOPTS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "hdc", "iPartId", "iStateId", "pszText", "cchText", "dwTextFlags", "pRect", "pOptions"]),
        #
        'GetThemeBitmap': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="GET_THEME_BITMAP_FLAGS"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "dwFlags", "phBitmap"]),
        #
        'GetThemeStream': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateId", "iPropId", "ppvStream", "pcbStream", "hInst"]),
        #
        'BufferedPaintInit': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'BufferedPaintUnInit': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'BeginBufferedPaint': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypeInt(signed=False, label="BP_BUFFERFORMAT"), SimTypePointer(SimTypeRef("BP_PAINTPARAMS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hdcTarget", "prcTarget", "dwFormat", "pPaintParams", "phdc"]),
        #
        'EndBufferedPaint': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hBufferedPaint", "fUpdateTarget"]),
        #
        'GetBufferedPaintTargetRect': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hBufferedPaint", "prc"]),
        #
        'GetBufferedPaintTargetDC': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hBufferedPaint"]),
        #
        'GetBufferedPaintDC': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hBufferedPaint"]),
        #
        'GetBufferedPaintBits': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeRef("RGBQUAD", SimStruct), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hBufferedPaint", "ppbBuffer", "pcxRow"]),
        #
        'BufferedPaintClear': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hBufferedPaint", "prc"]),
        #
        'BufferedPaintSetAlpha': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["hBufferedPaint", "prc", "alpha"]),
        #
        'BufferedPaintStopAllAnimations': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd"]),
        #
        'BeginBufferedAnimation': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("RECT", SimStruct), offset=0), SimTypeInt(signed=False, label="BP_BUFFERFORMAT"), SimTypePointer(SimTypeRef("BP_PAINTPARAMS", SimStruct), offset=0), SimTypePointer(SimTypeRef("BP_ANIMATIONPARAMS", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd", "hdcTarget", "prcTarget", "dwFormat", "pPaintParams", "pAnimationParams", "phdcFrom", "phdcTo"]),
        #
        'EndBufferedAnimation': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hbpAnimation", "fUpdateTarget"]),
        #
        'BufferedPaintRenderAnimation': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwnd", "hdcTarget"]),
        #
        'IsCompositionActive': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'GetThemeTransitionDuration': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hTheme", "iPartId", "iStateIdFrom", "iStateIdTo", "iPropId", "pdwDuration"]),
        #
        'OpenThemeDataForDpi': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd", "pszClassList", "dpi"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=unused-argument
from __future__ import annotations
from typing import cast, TYPE_CHECKING

import claripy
import pyvex
from archinfo.arch_arm import is_arm_arch

from angr.block import Block
from angr.errors import SimMemoryMissingError
from angr.calling_conventions import SimRegArg, SimStackArg, default_cc
from angr.engines.vex.claripy.datalayer import value as claripy_value
from angr.engines.light import SimEngineNostmtVEX
from angr.knowledge_plugins import Function
from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from angr.analyses.typehoon import typevars, typeconsts
from .engine_base import SimEngineVRBase, RichR
from .irsb_scanner import VEXIRSBScanner

if TYPE_CHECKING:
    pass

binop_handler = SimEngineNostmtVEX[
    "VariableRecoveryFastState", RichR[claripy.ast.BV | claripy.ast.FP], None
].binop_handler


class SimEngineVRVEX(
    SimEngineNostmtVEX["VariableRecoveryFastState", RichR[claripy.ast.BV | claripy.ast.FP], None],
    SimEngineVRBase["VariableRecoveryFastState", Block],
):
    """
    Implements the VEX engine for variable recovery analysis.
    """

    reg_read_stmts_to_ignore: set[int]
    stmts_to_lower: set[int]

    def __init__(self, *args, call_info=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.call_info = call_info or {}

    # Statement handlers

    def _is_top(self, expr: RichR) -> bool:
        return self.state.is_top(expr)

    def _top(self, bits: int) -> RichR[claripy.ast.BV]:
        return RichR(self.state.top(bits))

    def _process_block(self, whitelist=None):
        scanner = VEXIRSBScanner(self.project, logger=self.l)
        scanner._process(None, block=self.block)
        self.stmts_to_lower = scanner.stmts_to_lower
        self.reg_read_stmts_to_ignore = scanner.reg_read_stmts_to_ignore

        return super()._process_block(whitelist=whitelist)

    def _handle_stmt_WrTmp(self, stmt):
        self.tmps[stmt.tmp] = self._expr(stmt.data)

    def _handle_stmt_Put(self, stmt):
        offset = stmt.offset
        r = self._expr(stmt.data)
        size = stmt.data.result_size(self.tyenv) // 8

        if offset == self.arch.ip_offset:
            return
        self._assign_to_register(offset, r, size)

    def _handle_stmt_Store(self, stmt):
        addr_r = self._expr_bv(stmt.addr)
        size = stmt.data.result_size(self.tyenv) // 8
        r = self._expr(stmt.data)

        self._store(addr_r, r, size, stmt=stmt)

    def _handle_stmt_StoreG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:
            addr = self._expr_bv(stmt.addr)
            size = stmt.data.result_size(self.tyenv) // 8
            data = self._expr(stmt.data)
            self._store(addr, data, size, stmt=stmt)

    def _handle_stmt_LoadG(self, stmt):
        guard = self._expr(stmt.guard)
        if guard is True:
            addr = self._expr_bv(stmt.addr)
            if addr is not None:
                self.tmps[stmt.dst] = self._load(addr, self.tyenv.sizeof(stmt.dst) // 8)
        elif guard is False:
            data = self._expr(stmt.alt)
            self.tmps[stmt.dst] = data
        else:
            self.tmps[stmt.dst] = self._top(pyvex.get_type_size(self.tyenv.lookup(stmt.dst)))

    def _handle_stmt_LLSC(self, stmt: pyvex.IRStmt.LLSC):
        if stmt.storedata is None:
            # load-link
            addr = self._expr_bv(stmt.addr)
            size = self.tyenv.sizeof(stmt.result) // self.arch.byte_width
            data = self._load(addr, size)
            self.tmps[stmt.result] = data
        else:
            # store-conditional
            assert isinstance(stmt.storedata, pyvex.expr.RdTmp)
            storedata = self._expr(stmt.storedata)
            addr = self._expr_bv(stmt.addr)
            size = self.tyenv.sizeof(stmt.storedata.tmp) // self.arch.byte_width

            self._store(addr, storedata, size)

            result_size = self.tyenv.sizeof(stmt.result)
            self.tmps[stmt.result] = RichR(claripy.BVV(1, result_size))

    # Expression handlers

    def _expr_bv(self, expr) -> RichR[claripy.ast.BV]:
        result = self._expr(expr)
        assert isinstance(result.data, claripy.ast.BV)
        return cast(RichR[claripy.ast.BV], result)

    def _expr_fp(self, expr) -> RichR[claripy.ast.FP]:
        result = self._expr(expr)
        assert isinstance(result.data, claripy.ast.FP)
        return cast(RichR[claripy.ast.FP], result)

    def _handle_expr_Get(self, expr):
        reg_offset = expr.offset
        reg_size = expr.result_size(self.tyenv) // 8

        # because of how VEX implements MOVCC and MOVCS instructions in ARM THUMB mode, we need to skip the register
        # read if the immediate next instruction is an WrTmp(ITE).
        #
        # MOVCC           R3, #0
        #
        #    46 | ------ IMark(0xfeca2, 2, 1) ------
        #    47 | t299 = CmpLT32U(t8,0x00010000)
        #    48 | t143 = GET:I32(r3)      <-   this read does not exist
        #    49 | t300 = ITE(t299,0x00000000,t143)
        #    50 | PUT(r3) = t300
        #    51 | PUT(pc) = 0x000feca5
        if is_arm_arch(self.arch) and (self.ins_addr & 1) == 1 and self.stmt_idx < len(self.block.vex.statements) - 1:
            next_stmt = self.block.vex.statements[self.stmt_idx + 1]
            if isinstance(next_stmt, pyvex.IRStmt.WrTmp) and isinstance(next_stmt.data, pyvex.IRExpr.ITE):
                return RichR(self.state.top(reg_size * 8))

        force_variable_size = None
        if self.stmts_to_lower and self.stmt_idx in self.stmts_to_lower and reg_size == 8:
            force_variable_size = 4

        return self._read_from_register(
            reg_offset,
            reg_size,
            expr=expr,
            force_variable_size=force_variable_size,
            create_variable=self.stmt_idx not in self.reg_read_stmts_to_ignore,
        )

    def _handle_expr_GetI(self, expr):
        return self._top(expr.result_size(self.tyenv))

    def _handle_expr_ITE(self, expr):
        return self._top(expr.result_size(self.tyenv))

    def _handle_expr_GSPTR(self, expr):
        return self._top(expr.result_size(self.tyenv))

    def _handle_expr_VECRET(self, expr):
        return self._top(expr.result_size(self.tyenv))

    def _handle_expr_Load(self, expr: pyvex.IRExpr.Load) -> RichR:
        addr = self._expr_bv(expr.addr)
        size = expr.result_size(self.tyenv) // 8

        return self._load(addr, size)

    def _handle_expr_CCall(self, expr):  # pylint:disable=useless-return
        # ccalls don't matter
        return RichR(self.state.top(expr.result_size(self.tyenv)))

    def _handle_conversion(self, from_size, to_size, signed, operand) -> RichR:
        _ = self._expr(operand)
        return RichR(self.state.top(to_size))

    # Function handlers

    def _handle_function_concrete(self, func: Function):
        if func.prototype is None or func.calling_convention is None:
            return

        try:
            arg_locs = func.calling_convention.arg_locs(func.prototype)
        except (TypeError, ValueError):
            func.prototype = None
            return

        if None in arg_locs:
            return

        for arg_loc in arg_locs:
            for loc in arg_loc.get_footprint():
                if isinstance(loc, SimRegArg):
                    self._read_from_register(self.arch.registers[loc.reg_name][0] + loc.reg_offset, loc.size)
                elif isinstance(loc, SimStackArg):
                    try:
                        sp: MultiValues = self.state.register_region.load(self.arch.sp_offset, self.arch.bytes)
                    except SimMemoryMissingError:
                        pass
                    else:
                        one_sp = sp.one_value()
                        if one_sp is not None:
                            addr = RichR(loc.stack_offset + one_sp)
                            self._load(addr, loc.size)

        # clobber caller-saved registers
        for reg_name in func.calling_convention.CALLER_SAVED_REGS:
            reg_offset, reg_size = self.arch.registers[reg_name]
            self._assign_to_register(reg_offset, self._top(reg_size * self.arch.byte_width), reg_size)

    def _process_block_end(self, stmt_result, whitelist):
        # handles block-end calls
        current_addr = self.state.block_addr
        for target_func in self.call_info.get(current_addr, []):
            self._handle_function_concrete(target_func)

        if self.block.vex.jumpkind == "Ijk_Call":
            # emulates return values from calls
            cc = None
            for target_func in self.call_info.get(self.state.block_addr, []):
                if target_func.calling_convention is not None:
                    cc = target_func.calling_convention
                    break
            if cc is None:
                cc = default_cc(self.arch.name, platform=self.project.simos.name)(self.arch)
            if isinstance(cc.RETURN_VAL, SimRegArg):
                reg_offset, reg_size = self.arch.registers[cc.RETURN_VAL.reg_name]
                data = self._top(reg_size * self.arch.byte_width)
                self._assign_to_register(reg_offset, data, reg_size, create_variable=False)

        elif self.block.vex.jumpkind == "Ijk_Ret":
            # handles return statements

            # determine the size of the return register
            # TODO: Handle multiple return registers
            cc = self.state.function.calling_convention
            if cc is None:
                cc_cls = default_cc(self.arch.name, platform=self.project.simos.name)
                assert cc_cls is not None
                cc = cc_cls(self.arch)
            if isinstance(cc.RETURN_VAL, SimRegArg):
                ret_val_size = 0
                reg_offset = cc.RETURN_VAL.check_offset(self.arch)
                for i in range(cc.RETURN_VAL.size):
                    try:
                        _ = self.state.register_region.load(reg_offset + i, 1)
                        ret_val_size = i + 1
                    except SimMemoryMissingError:
                        break
                self.state.ret_val_size = (
                    ret_val_size if self.state.ret_val_size is None else max(self.state.ret_val_size, ret_val_size)
                )

    def _handle_expr_Const(self, expr):
        return RichR(
            claripy_value(expr.con.type, expr.con.value, size=expr.con.size), typevar=typeconsts.int_type(expr.con.size)
        )

    def _handle_expr_RdTmp(self, expr):
        try:
            return self.tmps[expr.tmp]
        except KeyError:
            return self._top(expr.result_size(self.tyenv))

    def _expr_pair(
        self, arg0: pyvex.expr.IRExpr, arg1: pyvex.expr.IRExpr
    ) -> tuple[RichR[claripy.ast.BV], RichR[claripy.ast.BV]] | tuple[RichR[claripy.ast.FP], RichR[claripy.ast.FP]]:
        r0 = self._expr(arg0)
        r1 = self._expr(arg1)
        assert type(r0) is type(r1)
        return r0, r1  # type: ignore

    @binop_handler
    def _handle_binop_Add(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])
        sum_ = r0.data + r1.data  # type: ignore

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(sum_, typevar=typeconsts.int_type(result_size), type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.AddN(r1.data.concrete_value))

        tc: set[typevars.TypeConstraint] = set()
        if r0.typevar is not None and r1.typevar is not None:
            tc.add(typevars.Subtype(r0.typevar, r1.typevar))
        return RichR(
            sum_,
            typevar=typevar,
            type_constraints=tc,
        )

    @binop_handler
    def _handle_binop_Sub(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])
        diff = r0.data - r1.data  # type: ignore

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(diff, typevar=typeconsts.int_type(result_size), type_constraints=None)

        typevar = None
        if r0.typevar is not None and r1.data.concrete:
            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data.concrete_value))

        return RichR(
            diff,
            typevar=typevar,
        )

    @binop_handler
    def _handle_binop_And(self, expr):
        r0 = self._expr_bv(expr.args[0])
        r1 = self._expr_bv(expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data & r1.data)

        if self.state.is_stack_address(r0.data):
            r = r0.data
        elif self.state.is_stack_address(r1.data):
            r = r1.data
        else:
            result_size = expr.result_size(self.tyenv)
            r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Xor(self, expr):
        r0 = self._expr_bv(expr.args[0])
        r1 = self._expr_bv(expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data ^ r1.data)

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Or(self, expr):
        r0 = self._expr_bv(expr.args[0])
        r1 = self._expr_bv(expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(r0.data | r1.data)

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Not(self, expr):
        arg = expr.args[0]
        r0 = self._expr_bv(arg)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete:
            # constants
            return RichR(~r0.data)

        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Mul(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            mul = r0.data * r1.data  # type: ignore
            return RichR(mul)

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_MullS(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            xt = r0.data.size()
            mul = r0.data.sign_extend(xt) * r1.data.sign_extend(xt)  # type: ignore
            return RichR(mul)

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_MullU(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            xt = r0.data.size()
            mul = r0.data.zero_extend(xt) * r1.data.zero_extend(xt)  # type: ignore
            return RichR(mul)

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_DivMod(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                signed = "U" in expr.op  # Iop_DivModU64to32 vs Iop_DivMod
                from_size = r0.data.size()
                to_size = r1.data.size()
                if signed:
                    quotient = r0.data.SDiv(claripy.SignExt(from_size - to_size, r1.data))
                    remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient),
                    )
                else:
                    quotient = r0.data // claripy.ZeroExt(from_size - to_size, r1.data)
                    remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)
                    quotient_size = to_size
                    remainder_size = to_size
                    result = claripy.Concat(
                        claripy.Extract(remainder_size - 1, 0, remainder),
                        claripy.Extract(quotient_size - 1, 0, quotient),
                    )

                return RichR(result)
            except ZeroDivisionError:
                pass

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Div(self, expr):
        r0, r1 = self._expr_pair(expr.args[0], expr.args[1])

        if r0.data.concrete and r1.data.concrete:
            # constants
            try:
                div = r0.data / r1.data  # type: ignore
                return RichR(div)
            except ZeroDivisionError:
                pass

        result_size = expr.result_size(self.tyenv)
        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Mod(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete and r1.data.concrete_value != 0:
            # constants
            try:
                if result_size != r1.data.size():
                    remainder = r0.data.SMod(claripy.SignExt(result_size - r1.data.size(), r1.data))
                else:
                    remainder = r0.data.SMod(r1.data)
                return RichR(remainder)
            except ZeroDivisionError:
                pass

        r = self.state.top(result_size)
        return RichR(r)

    @binop_handler
    def _handle_binop_Shr(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                claripy.LShR(r0.data, r1.data.concrete_value),
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    @binop_handler
    def _handle_binop_Sar(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                r0.data >> r1.data.concrete_value,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(r, typevar=r0.typevar)

    @binop_handler
    def _handle_binop_Shl(self, expr):
        arg0, arg1 = expr.args
        r0 = self._expr_bv(arg0)
        r1 = self._expr_bv(arg1)

        result_size = expr.result_size(self.tyenv)
        if r0.data.concrete and r1.data.concrete:
            # constants
            return RichR(
                r0.data << r1.data.concrete_value,
                typevar=typeconsts.int_type(result_size),
                type_constraints=None,
            )

        r = self.state.top(result_size)
        return RichR(
            r,
            typevar=r0.typevar,
        )

    @binop_handler
    def _handle_binop_CmpEQ(self, expr):
        arg0, arg1 = expr.args
        self._expr(arg0)
        self._expr(arg1)

        return RichR(self.state.top(1))

    _handle_binop_CmpNE = _handle_binop_CmpEQ
    _handle_binop_CmpLE = _handle_binop_CmpEQ
    _handle_binop_CmpLT = _handle_binop_CmpEQ
    _handle_binop_CmpGE = _handle_binop_CmpEQ
    _handle_binop_CmpGT = _handle_binop_CmpEQ

    def _handle_ExpCmpNE64(self, expr):
        _, _ = self._expr(expr.args[0]), self._expr(expr.args[1])
        return RichR(self.state.top(expr.result_size(self.tyenv)))



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("mgmtapi.dll")
prototypes = \
    {
        #
        'SnmpMgrOpen': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["lpAgentAddress", "lpAgentCommunity", "nTimeOut", "nRetries"]),
        #
        'SnmpMgrCtl': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["session", "dwCtlCode", "lpvInBuffer", "cbInBuffer", "lpvOUTBuffer", "cbOUTBuffer", "lpcbBytesReturned"]),
        #
        'SnmpMgrClose': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["session"]),
        #
        'SnmpMgrRequest': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeRef("SnmpVarBindList", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="SNMP_ERROR_STATUS"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["session", "requestType", "variableBindings", "errorStatus", "errorIndex"]),
        #
        'SnmpMgrStrToOid': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeRef("AsnObjectIdentifier", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["string", "oid"]),
        #
        'SnmpMgrOidToStr': SimTypeFunction([SimTypePointer(SimTypeRef("AsnObjectIdentifier", SimStruct), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["oid", "string"]),
        #
        'SnmpMgrTrapListen': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["phTrapAvailable"]),
        #
        'SnmpMgrGetTrap': SimTypeFunction([SimTypePointer(SimTypeRef("AsnObjectIdentifier", SimStruct), offset=0), SimTypePointer(SimTypeRef("AsnOctetString", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="SNMP_GENERICTRAP"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("SnmpVarBindList", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["enterprise", "IPAddress", "genericTrap", "specificTrap", "timeStamp", "variableBindings"]),
        #
        'SnmpMgrGetTrapEx': SimTypeFunction([SimTypePointer(SimTypeRef("AsnObjectIdentifier", SimStruct), offset=0), SimTypePointer(SimTypeRef("AsnOctetString", SimStruct), offset=0), SimTypePointer(SimTypeRef("AsnOctetString", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="SNMP_GENERICTRAP"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeRef("AsnOctetString", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("SnmpVarBindList", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["enterprise", "agentAddress", "sourceAddress", "genericTrap", "specificTrap", "community", "timeStamp", "variableBindings"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
import angr


class std____throw_logic_error(angr.SimProcedure):  # pylint:disable=redefined-builtin
    # pylint:disable=arguments-differ

    NO_RET = True
    ALT_NAMES = ("std::__throw_logic_error(char const*)",)

    def run(self, error):  # pylint:disable=unused-argument
        # FIXME: we need the concept of C++ exceptions to implement this right
        self.exit(1)



from __future__ import annotations
import warnings


once_set = set()


def once(key):
    if key in once_set:
        return False
    once_set.add(key)
    return True


already_complained = set()


def deprecated(replacement=None):
    def outer(func):
        def inner(*args, **kwargs):
            if func not in already_complained:
                if replacement is None:
                    warnings.warn(f"Don't use {func.__name__}", DeprecationWarning, stacklevel=2)
                else:
                    warnings.warn(f"Use {replacement} instead of {func.__name__}", DeprecationWarning, stacklevel=2)
                already_complained.add(func)
            return func(*args, **kwargs)

        return inner

    return outer



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("gdi32.dll")
prototypes = \
    {
        #
        'D3DKMTCreateAllocation': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEALLOCATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateAllocation2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEALLOCATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryResourceInfo': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYRESOURCEINFO", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryResourceInfoFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYRESOURCEINFOFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTShareObjects': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeRef("OBJECT_ATTRIBUTES", SimStruct), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cObjects", "hObjects", "pObjectAttributes", "dwDesiredAccess", "phSharedNtHandle"]),
        #
        'D3DKMTOpenNtHandleFromName': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENNTHANDLEFROMNAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenResourceFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENRESOURCEFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenSyncObjectFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENSYNCOBJECTFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenResource': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENRESOURCE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenResource2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENRESOURCE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyAllocation': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYALLOCATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyAllocation2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYALLOCATION2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetAllocationPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETALLOCATIONPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryAllocationResidency': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYALLOCATIONRESIDENCY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateDevice': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEDEVICE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyDevice': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYDEVICE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateContext': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATECONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyContext': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYCONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateSynchronizationObject': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATESYNCHRONIZATIONOBJECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateSynchronizationObject2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATESYNCHRONIZATIONOBJECT2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenSynchronizationObject': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENSYNCHRONIZATIONOBJECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroySynchronizationObject': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYSYNCHRONIZATIONOBJECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForSynchronizationObject': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORSYNCHRONIZATIONOBJECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForSynchronizationObject2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORSYNCHRONIZATIONOBJECT2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSignalSynchronizationObject': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SIGNALSYNCHRONIZATIONOBJECT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSignalSynchronizationObject2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SIGNALSYNCHRONIZATIONOBJECT2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTLock': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_LOCK", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUnlock': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_UNLOCK", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetDisplayModeList': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETDISPLAYMODELIST", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetDisplayMode': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETDISPLAYMODE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetMultisampleMethodList': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETMULTISAMPLEMETHODLIST", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPresent': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_PRESENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTRender': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_RENDER", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetRuntimeData': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETRUNTIMEDATA", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryAdapterInfo': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYADAPTERINFO", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenAdapterFromHdc': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENADAPTERFROMHDC", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenAdapterFromGdiDisplayName': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENADAPTERFROMGDIDISPLAYNAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenAdapterFromDeviceName': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENADAPTERFROMDEVICENAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCloseAdapter': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CLOSEADAPTER", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetSharedPrimaryHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETSHAREDPRIMARYHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTEscape': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ESCAPE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryStatistics': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYSTATISTICS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetVidPnSourceOwner': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETVIDPNSOURCEOWNER", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetPresentHistory': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETPRESENTHISTORY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetPresentQueueEvent': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hAdapter", "param1"]),
        #
        'D3DKMTCreateOverlay': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEOVERLAY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUpdateOverlay': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_UPDATEOVERLAY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTFlipOverlay': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_FLIPOVERLAY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyOverlay': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYOVERLAY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForVerticalBlankEvent': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORVERTICALBLANKEVENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetGammaRamp': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETGAMMARAMP", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetDeviceState': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETDEVICESTATE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateDCFromMemory': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEDCFROMMEMORY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyDCFromMemory': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYDCFROMMEMORY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetContextSchedulingPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETCONTEXTSCHEDULINGPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetContextSchedulingPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETCONTEXTSCHEDULINGPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetProcessSchedulingPriorityClass': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="D3DKMT_SCHEDULINGPRIORITYCLASS")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]),
        #
        'D3DKMTGetProcessSchedulingPriorityClass': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="D3DKMT_SCHEDULINGPRIORITYCLASS"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1"]),
        #
        'D3DKMTReleaseProcessVidPnSourceOwners': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetScanLine': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETSCANLINE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTChangeSurfacePointer': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHANGESURFACEPOINTER", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetQueuedLimit': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETQUEUEDLIMIT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPollDisplayChildren': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_POLLDISPLAYCHILDREN", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTInvalidateActiveVidPn': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_INVALIDATEACTIVEVIDPN", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckOcclusion': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKOCCLUSION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForIdle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORIDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckMonitorPowerState': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKMONITORPOWERSTATE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckExclusiveOwnership': SimTypeFunction([], SimTypeChar(label="Byte")),
        #
        'D3DKMTCheckVidPnExclusiveOwnership': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKVIDPNEXCLUSIVEOWNERSHIP", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetDisplayPrivateDriverFormat': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETDISPLAYPRIVATEDRIVERFORMAT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSharedPrimaryLockNotification': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SHAREDPRIMARYLOCKNOTIFICATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSharedPrimaryUnLockNotification': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SHAREDPRIMARYUNLOCKNOTIFICATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateKeyedMutex': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEKEYEDMUTEX", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenKeyedMutex': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENKEYEDMUTEX", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyKeyedMutex': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYKEYEDMUTEX", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTAcquireKeyedMutex': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ACQUIREKEYEDMUTEX", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTReleaseKeyedMutex': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_RELEASEKEYEDMUTEX", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateKeyedMutex2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEKEYEDMUTEX2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenKeyedMutex2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENKEYEDMUTEX2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTAcquireKeyedMutex2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ACQUIREKEYEDMUTEX2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTReleaseKeyedMutex2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_RELEASEKEYEDMUTEX2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTConfigureSharedResource': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CONFIGURESHAREDRESOURCE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetOverlayState': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETOVERLAYSTATE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckSharedResourceAccess': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKSHAREDRESOURCEACCESS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOfferAllocations': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OFFERALLOCATIONS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTReclaimAllocations': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_RECLAIMALLOCATIONS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateOutputDupl': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATE_OUTPUTDUPL", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyOutputDupl': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROY_OUTPUTDUPL", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOutputDuplGetFrameInfo': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OUTPUTDUPL_GET_FRAMEINFO", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOutputDuplGetMetaData': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OUTPUTDUPL_METADATA", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOutputDuplGetPointerShapeData': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OUTPUTDUPL_GET_POINTER_SHAPE_DATA", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOutputDuplReleaseFrame': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OUTPUTDUPL_RELEASE_FRAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOutputDuplPresent': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OUTPUTDUPLPRESENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTEnumAdapters': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ENUMADAPTERS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTEnumAdapters2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ENUMADAPTERS2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenAdapterFromLuid': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENADAPTERFROMLUID", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryRemoteVidPnSourceFromGdiDisplayName': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYREMOTEVIDPNSOURCEFROMGDIDISPLAYNAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetVidPnSourceOwner1': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETVIDPNSOURCEOWNER1", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForVerticalBlankEvent2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORVERTICALBLANKEVENT2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetSyncRefreshCountWaitTarget': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETSYNCREFRESHCOUNTWAITTARGET", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetDWMVerticalBlankEvent': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETVERTICALBLANKEVENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPresentMultiPlaneOverlay': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_PRESENT_MULTIPLANE_OVERLAY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetSharedResourceAdapterLuid': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETSHAREDRESOURCEADAPTERLUID", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckMultiPlaneOverlaySupport': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKMULTIPLANEOVERLAYSUPPORT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetContextInProcessSchedulingPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETCONTEXTINPROCESSSCHEDULINGPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetContextInProcessSchedulingPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETCONTEXTINPROCESSSCHEDULINGPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTMakeResident': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_MAKERESIDENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTEvict': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_EVICT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForSynchronizationObjectFromCpu': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORSYNCHRONIZATIONOBJECTFROMCPU", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSignalSynchronizationObjectFromCpu': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SIGNALSYNCHRONIZATIONOBJECTFROMCPU", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTWaitForSynchronizationObjectFromGpu': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_WAITFORSYNCHRONIZATIONOBJECTFROMGPU", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSignalSynchronizationObjectFromGpu': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SIGNALSYNCHRONIZATIONOBJECTFROMGPU", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSignalSynchronizationObjectFromGpu2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SIGNALSYNCHRONIZATIONOBJECTFROMGPU2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreatePagingQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEPAGINGQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyPagingQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_DESTROYPAGINGQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTLock2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_LOCK2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUnlock2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_UNLOCK2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTInvalidateCache': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_INVALIDATECACHE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTMapGpuVirtualAddress': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_MAPGPUVIRTUALADDRESS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTReserveGpuVirtualAddress': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_RESERVEGPUVIRTUALADDRESS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTFreeGpuVirtualAddress': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_FREEGPUVIRTUALADDRESS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUpdateGpuVirtualAddress': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_UPDATEGPUVIRTUALADDRESS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetResourcePresentPrivateDriverData': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_GETRESOURCEPRESENTPRIVATEDRIVERDATA", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateContextVirtual': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATECONTEXTVIRTUAL", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSubmitCommand': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SUBMITCOMMAND", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenSyncObjectFromNtHandle2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENSYNCOBJECTFROMNTHANDLE2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenSyncObjectNtHandleFromName': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENSYNCOBJECTNTHANDLEFROMNAME", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryVideoMemoryInfo': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYVIDEOMEMORYINFO", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTChangeVideoMemoryReservation': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHANGEVIDEOMEMORYRESERVATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTRegisterTrimNotification': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_REGISTERTRIMNOTIFICATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUnregisterTrimNotification': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_UNREGISTERTRIMNOTIFICATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckMultiPlaneOverlaySupport2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKMULTIPLANEOVERLAYSUPPORT2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPresentMultiPlaneOverlay2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_PRESENT_MULTIPLANE_OVERLAY2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTReclaimAllocations2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_RECLAIMALLOCATIONS2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetStablePowerState': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETSTABLEPOWERSTATE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryClockCalibration': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYCLOCKCALIBRATION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryVidPnExclusiveOwnership': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYVIDPNEXCLUSIVEOWNERSHIP", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTAdjustFullscreenGamma': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_ADJUSTFULLSCREENGAMMA", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetVidPnSourceHwProtection': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETVIDPNSOURCEHWPROTECTION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTMarkDeviceAsError': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_MARKDEVICEASERROR", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTFlushHeapTransitions': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_FLUSHHEAPTRANSITIONS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetHwProtectionTeardownRecovery': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETHWPROTECTIONTEARDOWNRECOVERY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryProcessOfferInfo': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYPROCESSOFFERINFO", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTTrimProcessCommitment': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_TRIMPROCESSCOMMITMENT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTUpdateAllocationProperty': SimTypeFunction([SimTypePointer(SimTypeRef("D3DDDI_UPDATEALLOCPROPERTY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCheckMultiPlaneOverlaySupport3': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CHECKMULTIPLANEOVERLAYSUPPORT3", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPresentMultiPlaneOverlay3': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_PRESENT_MULTIPLANE_OVERLAY3", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetFSEBlock': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETFSEBLOCK", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryFSEBlock': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYFSEBLOCK", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateHwContext': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEHWCONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyHwContext': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYHWCONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateHwQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEHWQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyHwQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYHWQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSubmitCommandToHwQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SUBMITCOMMANDTOHWQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSubmitWaitForSyncObjectsToHwQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SUBMITWAITFORSYNCOBJECTSTOHWQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSubmitSignalSyncObjectsToHwQueue': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SUBMITSIGNALSYNCOBJECTSTOHWQUEUE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetAllocationPriority': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETALLOCATIONPRIORITY", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetMultiPlaneOverlayCaps': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GET_MULTIPLANE_OVERLAY_CAPS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetPostCompositionCaps': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GET_POST_COMPOSITION_CAPS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTPresentRedirected': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_PRESENT_REDIRECTED", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetVidPnSourceOwner2': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SETVIDPNSOURCEOWNER2", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTSetMonitorColorSpaceTransform': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_SET_COLORSPACE_TRANSFORM", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCreateProtectedSession': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CREATEPROTECTEDSESSION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTDestroyProtectedSession': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_DESTROYPROTECTEDSESSION", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryProtectedSessionStatus': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYPROTECTEDSESSIONSTATUS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTQueryProtectedSessionInfoFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_QUERYPROTECTEDSESSIONINFOFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenProtectedSessionFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENPROTECTEDSESSIONFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTGetProcessDeviceRemovalSupport': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_GETPROCESSDEVICEREMOVALSUPPORT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTOpenKeyedMutexFromNtHandle': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_OPENKEYEDMUTEXFROMNTHANDLE", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTRegisterVailProcess': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
        #
        'D3DKMTCancelPresents': SimTypeFunction([SimTypePointer(SimTypeRef("D3DKMT_CANCEL_PRESENTS", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
import logging
from enum import Enum, auto
from typing import TYPE_CHECKING
from dataclasses import dataclass

from angr.analyses import Analysis, AnalysesHub


if TYPE_CHECKING:
    from angr.knowledge_plugins import Function


log = logging.getLogger(__name__)


class CodeCaveClassification(Enum):
    """
    Type of code caves.
    """

    ALIGNMENT = auto()
    UNREACHABLE = auto()


@dataclass
class CodeCave:
    """
    Describes a code cave in a binary.
    """

    func: Function | None
    addr: int
    size: int
    classification: CodeCaveClassification


class CodeCaveAnalysis(Analysis):
    """
    Best-effort static location of potential vacant code caves for possible code injection:
    - Padding functions
    - Unreachable code
    """

    codecaves: list[CodeCave]

    def __init__(self):
        self.codecaves = []

        if len(self.project.kb.functions) == 0 and self.project.kb.cfgs.get_most_accurate() is None:
            log.warning("Please run CFGFast analysis first, to identify functions")
            return

        # Alignment functions
        for func in self.project.kb.functions.values():
            if func.is_alignment:
                for block in func.blocks:
                    self.codecaves.append(CodeCave(func, block.addr, block.size, CodeCaveClassification.ALIGNMENT))

        # Unreachable code
        for func in self.project.kb.functions.values():
            if func.is_alignment or func.is_plt or func.is_simprocedure or func.addr in self.project.kb.labels:
                continue

            in_degree = self.project.kb.callgraph.in_degree(func.addr)
            if in_degree == 0 or (
                in_degree == 1
                and self.project.kb.functions[next(self.project.kb.callgraph.predecessors(func.addr))].is_alignment
            ):
                for block in func.blocks:
                    self.codecaves.append(CodeCave(func, block.addr, block.size, CodeCaveClassification.UNREACHABLE))

            # FIXME: find dead blocks with argument propagation
            # FIXME: find dead blocks with external coverage info


AnalysesHub.register_default("CodeCaves", CodeCaveAnalysis)
